"""Init"""
import sys

from geochron._version import __version__  # noqa: F401
from geochron.utils.conditional_imports import ConditionalPackageInterceptor
from geochron.chronnet import convert_chronnet
from geochron.timehex import convert_timehex
from geochron.geotimehash import convert_geotimehash
from geochron.geosynchnet import convert_geosynchnet

ConditionalPackageInterceptor.permit_packages(
    {
        'networkx': 'networkx>=3.0,<4.0',
        'plotly': 'plotly>=5,<6',
        'timehash': 'timehash>=1.2,<2',
    }
)
sys.meta_path.append(ConditionalPackageInterceptor)  # type: ignore

__all__ = [
    'convert_chronnet',
    'convert_timehex',
    'convert_geotimehash',
    'convert_geosynchnet' 
]



""" Representation as time hexes """
from datetime import  timedelta
from typing import Callable, List
import pandas as pd
from geostructures.collections import FeatureCollection, Track
from geochron.time_slicing import get_timestamp_intervals, time_slice_track





def hash_tracks_into_timehexdf(track_list: List, timestamps: List, hash_func: Callable):
    """
    Converts a list of tracks into a pandas dataframe using
    a specified hashing function with intervals reflected
    in a corresponding timestamp list 
    
    Args:
        track_list: a list of tracks broken down by equal intervals

        timestamps: a list of corresponding timestamps

        hash_func: the hashing function

    Returns:
        A pandas dataframe
    """
    interval_start = track_list[0].start
    row_list = []
    for track,timestamp  in zip(track_list, timestamps):
        hashmap = hash_func(track)
        start_string= interval_start.strftime("%Y-%m-%d %H:%M:%S")
        end_string= timestamp.strftime("%Y-%m-%d %H:%M:%S")
        interval = start_string + ", " + end_string
        new_row = pd.Series(data=hashmap, name=interval)
        row_list.append(new_row)
        interval_start = timestamp

    df = pd.DataFrame(row_list)

    df = df.reset_index(names='interval')

    split_df = df['interval'].str.split(',', expand=True)# pylint: disable=unsubscriptable-object
    df.loc[:, 'start_time'] = split_df[0]
    df.loc[:, 'end_time'] = split_df[1]

    return df

def convert_timehex(fcol: FeatureCollection, time_delta: timedelta, hash_func: Callable):
    """
    Converts a FeatureCollection into a timehex representation with a specified time interval
    using a specified hashing function
    
    Args:
        fcol: a FeatureCollection with time bound shapes

        time_delta: the desired time interval

        hash_func: the hashing function

    Returns:
        A pandas dataframe
    """
    track = Track(fcol.geoshapes)

    timestamps = get_timestamp_intervals(track, time_delta)

    track_list = time_slice_track(track, timestamps)

    timehex_df = hash_tracks_into_timehexdf(track_list, timestamps, hash_func)


    return timehex_df



import datetime as dt
from geochron.timehex import hash_tracks_into_timehexdf, convert_timehex
from geostructures import Coordinate, GeoPoint
from geostructures.collections import  Track
from geostructures.geohash import H3Hasher

hasher = H3Hasher(resolution = 10)
test_timestamps = [dt.datetime(2020, 1, 1, 9, 5, tzinfo=dt.timezone.utc),
dt.datetime(2020, 1, 1, 10, 5, 1, tzinfo=dt.timezone.utc)]

def test_hash_tracks_into_timehexdf():

    track1 = Track(
    [
        GeoPoint(Coordinate(-0.104154, 51.511920), dt=dt.datetime(2020, 1, 1, 8, 5))
    ]
    )
    track2 = Track(
    [
        GeoPoint(Coordinate(-0.096533, 51.511903), dt=dt.datetime(2020, 1, 1, 9, 23)),
        GeoPoint(Coordinate(-0.083765, 51.514423), dt=dt.datetime(2020, 1, 1, 9, 44)),
        GeoPoint(Coordinate(-0.087478, 51.508595), dt=dt.datetime(2020, 1, 1, 10, 5)),
    ]
    )

    track_list = [track1, track2]
    test_df = hash_tracks_into_timehexdf(track_list, test_timestamps, hasher.hash_collection)

    assert test_df['8a194ad3056ffff'].values[1] == 1


def test_convert_chronnet():
    track = Track(
    [
        GeoPoint(Coordinate(-0.104154, 51.511920), dt=dt.datetime(2020, 1, 1, 8, 5)),
        GeoPoint(Coordinate(-0.096533, 51.511903), dt=dt.datetime(2020, 1, 1, 9, 23)),
        GeoPoint(Coordinate(-0.083765, 51.514423), dt=dt.datetime(2020, 1, 1, 9, 44)),
        GeoPoint(Coordinate(-0.087478, 51.508595), dt=dt.datetime(2020, 1, 1, 10, 5)),
    ]
    )

    test_timehex = convert_timehex(track, dt.timedelta(hours=1), hasher.hash_collection)

    assert test_timehex['8a194ad3056ffff'].values[1] == 1


""" Representation as geosynchnet"""
from datetime import  timedelta
from typing import Callable
import pandas as pd

from geostructures.collections import  FeatureCollection,Track
from geochron.chronnet import hash_tracks_into_netdf
from geochron.time_slicing import get_timestamp_intervals, time_slice_track

def geosynchnet_create(df: pd.DataFrame):
    """
    Converts a properly formatted pandas dataframe with the columns
    of cell and time to a networkx network where nodes are locations
    and edges are formed between nodes with that share a time. 
    
    Args:
        df: a pandas dataframe withe two columns cell and time

    Returns:
        A networkx network
    """
    # pylint: disable=import-outside-toplevel
    import networkx as nx # type: ignore
    # Initialize an empty network
    net = nx.Graph()

    # Group the dataframe by time
    grouped = df.groupby('time')

    # For each time group, add edges between all cells in the group
    for name, group in grouped:
        cells = group['cell'].tolist()
        for i in range(len(cells)):
            for j in range(i+1, len(cells)):
                net.add_edge(cells[i], cells[j])

    return net

def convert_geosynchnet(fcol: FeatureCollection, time_delta: timedelta,
     hash_func: Callable):
    """
    Converts a FeatureCollection into a chronnet with a specified time interval
    using a specified hashing function
    
    Args:
        fcol: a FeatureCollection with time bound shapes 

        time_delta: the desired time interval

        hash_func: the hashing function

    Returns:
        A networkx network 
    """
    track = Track(fcol.geoshapes)
    timestamps = get_timestamp_intervals(track, time_delta)

    track_list = time_slice_track(track, timestamps)

    df = hash_tracks_into_netdf(track_list, timestamps, hash_func)

    geosynchnet = geosynchnet_create(df)

    return geosynchnet



""" Functions needed to slice tracks by time"""
from datetime import  timedelta
from typing import List
from geostructures.collections import  Track



def get_timestamp_intervals(track: Track, time_delta: timedelta):
    """
    gets the timestamps for a Track partitioned by a specified length
    of time in hours  
    
    Args:
        track: the target geostructures Track

        time_delta: the desired time interval

    Returns:
        A list of timestamps starting from an interval from the start
    """
    start_time = track.start
    end_time = track.end
    timestamps = []

    # get timestamp intervals
    date_x = start_time
    while date_x < end_time:
        date_x += time_delta
        timestamps.append(date_x)

    # change the last value to be inclusive
    timestamps[-1] = timestamps[-1] + timedelta(seconds=1)


    return timestamps


def time_slice_track(track: Track, timestamps: List):
    """
    Slices a Track into several tracks that are partitioned by a list
    of time stamps  
    
    Args:
        track: the target geostructures Track

        timestamps= a list of timestamps

    Returns:
        A list of tracks
    """
    start_time = track.start
    sliced_tracks = []


    for timestamp in timestamps:
        sliced_track = track[start_time: timestamp]
        start_time = timestamp
        sliced_tracks.append(sliced_track)

    return sliced_tracks



from datetime import datetime, timedelta, timezone
from geochron.geotimehash import precision_delta, generate_times, timehash_geoshape,\
append_timehash_to_geohashmap, breakdown_hashmap_by_suffix, combine_dicts, convert_geotimehash
from geostructures import Coordinate, GeoCircle, GeoPoint
from geostructures.collections import  Track
from geostructures.geohash import H3Hasher
from geostructures.time import  TimeInterval

def test_precision_delta():
    delta1 = precision_delta(1)
    delta2 = precision_delta(2)
    delta3 = precision_delta(3)
    delta4 = precision_delta(4)
    delta5 = precision_delta(5)
    delta6 = precision_delta(6)
    delta7 = precision_delta(7)
    delta8 = precision_delta(8)
    delta9 = precision_delta(9)
    delta10 = precision_delta(10)

    assert delta1 == timedelta(days = 5840)
    assert delta2 == timedelta(days = 730)
    assert delta3 == timedelta(days = 91.2)
    assert delta4 == timedelta(days = 11.4)
    assert delta5 == timedelta(hours = 34.2)
    assert delta6 == timedelta(hours = 4.2)
    assert delta7 == timedelta(minutes = 32)
    assert delta8 == timedelta(minutes = 4)
    assert delta9 == timedelta(seconds = 30)
    assert delta10 == timedelta(seconds = 3.6)

def test_generate_times():
    start_time = datetime(2020, 1, 1, 9, 30, tzinfo=timezone.utc)
    end_time = datetime(2020, 1, 1, 9, 42, tzinfo=timezone.utc)
    time_list = generate_times(start_time, end_time, 8)
    time2 = datetime(2020, 1, 1, 9, 38, tzinfo=timezone.utc).timestamp()

    assert time_list[2] == time2

def test_timehash_geoshape():
    circle2 = GeoCircle(
    Coordinate(-0.118092, 51.509865), 
    radius=5000, 
    dt=TimeInterval(datetime(2020, 1, 1, 9, 30), datetime(2020, 1, 1, 9, 42))
    )
    timehash_list = timehash_geoshape(circle2, 8)

    assert timehash_list[2] == 'b0ffffbc'

def test_append_timehash_to_geohashmap():
    geohashmap = {'8b194ad32161fff': 1.0,
                '8b194ad32b23fff': 1.0,
                '8b194ad32b1dfff': 1.0,
                '8b194ad32b33fff': 1.0}
    
    timehash_list = ['b0ffffbc']

    geotime_map = append_timehash_to_geohashmap(geohashmap, timehash_list)

    assert geotime_map['8b194ad32161fff_b0ffffbc'] == 1

def test_breakdown_hashmap_by_suffix():
    hashmap = {'8b194ad32161fff_b0ffffbc': 1.0,
                '8b194ad32b23fff_b0ffffbc': 1.0,
                '8b194ad32b1dfff_b0ffffba': 1.0,
                '8b194ad32b33fff_b0ffffba': 1.0}
    
    dict_of_dicts = breakdown_hashmap_by_suffix(hashmap)

    assert dict_of_dicts['b0ffffbc']['8b194ad32b23fff_b0ffffbc'] == 1

def test_combine_dicts():
    testdict = {'a':{'r':2,'b':3},
                'b':{'m':2,'n':3}}
    result_dict = combine_dicts(testdict)

    assert result_dict['m'] == 2

def test_convert_geotimehash():
    hasher = H3Hasher(resolution = 10)
    track = Track(
    [
        GeoPoint(Coordinate(-0.104154, 51.511920), dt=datetime(2020, 1, 1, 8, 5)),
        GeoPoint(Coordinate(-0.096533, 51.511903), dt=datetime(2020, 1, 1, 9, 23)),
        GeoPoint(Coordinate(-0.083765, 51.514423), dt=datetime(2020, 1, 1, 9, 44)),
        GeoPoint(Coordinate(-0.087478, 51.508595), dt=datetime(2020, 1, 1, 10, 5)),
    ]
    )

    geotimehash = convert_geotimehash(track, 8, hasher.hash_collection)

    assert geotimehash['8a194ad3056ffff_b0ffffbe'] == 1

    
    


"""Package build script"""
import os
import re
import setuptools

ver_file = f'geochron{os.sep}_version.py'
__version__ = None

# Pull package version number from _version.py
with open(ver_file, 'r') as f:
    for line in f.readlines():
        if re.match(r'^\s*#', line):  # comment
            continue

        ver_line = line
        verstr = re.match(r"^.*=\s+'(v\d+\.\d+\.\d+(?:\.[a-zA-Z0-9]+)?)'", ver_line)
        if verstr is not None and len(verstr.groups()) == 1:
            __version__ = verstr.groups()[0]
            break

    if __version__ is None:
        raise EnvironmentError(f'Could not find valid version number in {ver_file}; aborting setup')

with open("./README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="geochron",
    version=__version__,
    author="Eli Talbert",
    author_email="",
    description="A companion package to geostructures enabling geo-spatial-temporal data structures.",
    long_description=long_description,
    long_description_content_type='text/markdown',
    url="https://github.com/etalbert/geochron",
    packages=setuptools.find_packages(
        include=('geochron*', ),
        exclude=('*tests', 'tests*')
    ),
    package_data={"geochron": ["py.typed"]},
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent"
    ],
    python_requires='>=3.8',
    install_requires=[
        'geostructures>= 0.8.1, <1.0',
        'numpy>=1,<2',  
        'pandas>=2,<3'
    ],
)